// Saeed Mahani
// Project 5 - 3D Spacial Interactive
// MAT 259 - Data Visualization
// W13

import processing.opengl.*;
import de.bezier.data.sql.*;
import peasy.*;

static final int COUNTRY_COLUMN = 1;
static final int LATITUDE_COLUMN = 5;
static final int LONGITUDE_COLUMN = 6;
static final int LOCATION_END_ROW = 242;
static final int LOCATION_ROW_OFFSET = 2;
static final float WORLD_RADIUS = 250;
static final float MAX_BAR_HEIGHT = 100;
static final float BAR_WIDTH = 9;
static final int COUNT_UPPER_LIMIT = 20000;
static final int YEAR_COUNT = 7;

static final int UNDEFINED = 0;
static final int NORTH_AMERICA = 1;
static final int SOUTH_AMERICA = 2;
static final int ASIA = 3;
static final int EUROPE = 4;
static final int AFRICA = 5;
static final int AUSTRALIA = 6;
static final int ANTARCTICA = 7;


PeasyCam cam;
MySQL msql;
PFont font;
float xRotation = 0;
float yRotation = 0;
float zRotation = 0;
int maxCount = COUNT_UPPER_LIMIT;
color labelColor, barColor, worldWireframeColor, positiveColor, negativeColor;
color[] continentColor = new color[8];
int visibleYear = 0;
boolean areCountryLabelsVisible = true;
float cam_angle_Y = 0;
float behind_cam_angle_Y = PI;
boolean isYearRangeMode = false;
int visibleYearRange = 0; // 0: 2006-2007, 1: 2007-2008, ..., 5: 2011-2012

ArrayList<Location> locations = new ArrayList<Location>(200); 

class Location {
  String country;
  float latitude;
  float longitude;
  int continent;
  int[] counts;
  color barColor;

  Location(String country, float latitude, float longitude, int continent) {
    this.country = country;
    this.latitude = latitude;
    this.longitude = longitude;
    this.continent = continent;
    this.counts = new int[YEAR_COUNT];

    // set bar color

    this.barColor = continentColor[continent];

  }
};


void setup() {
  size(1024, 768, OPENGL);
  perspective(PI/3.0, float(width)/height, 0.1, 5000);
  colorMode(HSB, 360, 100, 100);
  frameRate(60);
  smooth();

  // Fonts & colors

  labelColor = color(0, 0, 100, 200);
  barColor = color(117, 43, 83, 90);
  //barColor = color(0, 0, 20, 95);
  worldWireframeColor = color(0, 0, 60, 60);
  //positiveColor = color()

  final int sat = 57;
  final int bri = 93;

  continentColor[UNDEFINED]     = color(117, 0, 50, 90);
  continentColor[NORTH_AMERICA] = color(189, sat, bri, 90);
  continentColor[SOUTH_AMERICA] = color(117, sat, bri, 90);
  continentColor[ASIA]          = color(150, sat, bri, 90);
  continentColor[EUROPE]        = color(350, sat, bri, 90); 
  continentColor[AFRICA]        = color(80, sat, bri, 90);
  continentColor[AUSTRALIA]     = color(180, sat, bri, 90);
  continentColor[ANTARCTICA]    = color(117, 0, 50, 90);


  //// Setup the camera

  cam = new PeasyCam(this, 0, 0, 0, 800);
  cam.setMinimumDistance(400);
  cam.setMaximumDistance(1000);
  font = createFont("Helvetica", 40); 
  textFont(font);


  //// Load country codes (e.g: "US" -> "united states")

  HashMap<String, String> countryByCode = new HashMap<String, String>();          // key: country code, value: country name
  HashMap<String, Integer> continentByCode = new HashMap<String, Integer>();  // key: country code, value: continent enum

  String[] lines = loadStrings("countrycodes.csv");

  for (String line : lines){
    String[] cols = split(line, ',');
    String code = cols[0];
    String country = cols[1].replace("\'", "");
    int continent = int(cols[2]);
    continentByCode.put(code, continent);
    countryByCode.put(code, country);
  }


  //// Load locations

  lines = loadStrings("countrylatlon.csv");
  for (int lineNum = LOCATION_ROW_OFFSET; lineNum<LOCATION_END_ROW; lineNum++){
    String[] line = split(lines[lineNum], ',');
    String countryCode = line[COUNTRY_COLUMN].replace("\"", "");
    String country = countryByCode.get(countryCode);
    if (country == null) {
      continue;
    }
    int continent = continentByCode.get(countryCode);
    float latitude = float(line[LATITUDE_COLUMN]);
    float longitude = float(line[LONGITUDE_COLUMN]);
    locations.add(new Location(country, latitude, longitude, continent));
  }

  lines = loadStrings("countrycounts.csv");
  if (lines == null) {
    println("cache doesn't exist, performing query");
    performQuery();
  }
  else {
    for (int i = 0; i<YEAR_COUNT; i++){
      String[] locationCounts = split(lines[i+1], ',');
      for (int j = 0; j<locationCounts.length; j++){
        int count = int(locationCounts[j]);
        locations.get(j).counts[i] = count;
        //maxCount = max(maxCount, count);
      }
    }
  }


  
}


void draw() {

  background(0);
  textAlign(CENTER, CENTER);
  textSize(13);

  //// World sphere

  stroke(worldWireframeColor);
  noFill();
  sphereDetail(360/15, 180/15);
  sphere(WORLD_RADIUS);
  
  // Iterate over locations


  for (int i = 0; i<locations.size(); i++) {
  
    Location l = locations.get(i);
    float barHeight;
    if (isYearRangeMode)
      barHeight = min(float(l.counts[visibleYearRange+1]) / maxCount, 1.0) * MAX_BAR_HEIGHT;
    else 
      barHeight = min(float(l.counts[visibleYear]) / maxCount, 1.0) * MAX_BAR_HEIGHT;
    
    // Move to sphere surface at lat,lon position
    pushMatrix();
    float lonAngle = radians(l.longitude);
    rotateY(lonAngle); // Longitude rotation
    rotateX(radians(l.latitude)); // Latitude rotation
    translate(0, 0, WORLD_RADIUS+4);

      if (areCountryLabelsVisible) {

        // Rotate so labels stick out of sphere (at the tip of bars)
        pushMatrix();
        translate(0, 0, barHeight + 6);

        boolean shouldFlipLabel = false;

        if (cam_angle_Y > 0) {
          if (lonAngle > cam_angle_Y || lonAngle < behind_cam_angle_Y)
            shouldFlipLabel = true;
        }
        else {
          if (lonAngle > cam_angle_Y && lonAngle < behind_cam_angle_Y)
            shouldFlipLabel = true;
        }

        if (shouldFlipLabel ){
          rotateY(-PI/2.0);
          textAlign(LEFT, CENTER);
        }
        else {
          rotateY(PI/2.0);
          textAlign(RIGHT, CENTER);
        }

        // Draw location label
        fill(labelColor);
        textSize(map(barHeight, 0, MAX_BAR_HEIGHT, 9, 14));
        text(l.country, 0, 0);
        
        popMatrix();
      }    



    

    // Translate out half the depth of the bar (otherwise bar penetrates the sphere)
    translate(0, 0, barHeight/2.0);

    // Draw location count bar

    noStroke();

    if (isYearRangeMode) {
      int firstYear = l.counts[visibleYearRange];
      int secondYear = l.counts[visibleYearRange+1];
      float diff = float(secondYear - firstYear) / firstYear;

      int sat = int(min(1, abs(diff)/1.2)*100.0);

      if (diff > 0){
        fill(140, sat, sat, 90);
      }
      else {
        fill(350, sat, sat, 90); // sat 57
      }
    }
    else {
      fill(l.barColor);
    }

    box(BAR_WIDTH, BAR_WIDTH, barHeight);
    //ellipse(0,0,barHeight, barHeight);
    popMatrix();
  }

  
  // HUD

  cam.beginHUD();
  textAlign(LEFT, TOP);
  fill(300);
  textSize(15);
  //text("YEAR: " + (visibleYear+2006), 10, 10);
  //text("xRotation: " + xRotation + " yRotation: " + yRotation + " zRotation: " + zRotation, 10, 30);
  textAlign(RIGHT, TOP);
  textSize(38);

  if (isYearRangeMode) 
    text((visibleYearRange+2006) + " - " + (visibleYearRange+2007), width-10, 10);
  else
    text(visibleYear+2006, width-10, 10);
  cam.endHUD();
}


void keyPressed() {

  if (key == CODED) {

    if (isYearRangeMode) {
      if (keyCode == LEFT) {
        visibleYearRange = (visibleYearRange + YEAR_COUNT-2) % (YEAR_COUNT-1);
      } 
      else if (keyCode == RIGHT) {
        visibleYearRange = (visibleYearRange + 1) % (YEAR_COUNT-1);
      }
    }
    else {
      if (keyCode == LEFT) {
        visibleYear = (visibleYear + YEAR_COUNT-1) % YEAR_COUNT;
      } 
      else if (keyCode == RIGHT) {
        visibleYear = (visibleYear + 1) % YEAR_COUNT;
      }
    }

  }
  else {

    switch (key) {
      case '2' :
        areCountryLabelsVisible = !areCountryLabelsVisible;
      break;  

      case '1':
        isYearRangeMode = !isYearRangeMode;
      break;
    }

    if (key == 'r') {
      cam.reset(1000);
    } 

    if (key == 'a') {
      cam.rotateY(PI/2.0);
      //cam.lookAt(random(-100.0,50.0),random(-200.0,100.0),random(-100.0,300.0), 1000);
      //xRotationTarget = random(10.0);
      //yRotationTarget = random(1.0);
      //zRotationTarget = random(1.0);
    }



    if (key == 'c') {
      float[] pos = cam.getPosition();
      println("cam position: " + pos[0] + " " + pos[1] + " " + pos[2]);
    }
  }    

}



void mouseDragged() {
  float[] rotations = cam.getRotations();
  xRotation = rotations[0];
  yRotation = rotations[1];
  zRotation = rotations[2];   

  float[] pos = cam.getPosition();
  //float signed_angle = atan2(800, 0) - atan2(pos[2],pos[0]);
  cam_angle_Y = -atan2(0*pos[2] - 800.0*pos[0], 0*pos[0] + 800.0*pos[2]);
  behind_cam_angle_Y = cam_angle_Y > 0 ? (cam_angle_Y - PI) : (cam_angle_Y + PI);
  //println("signed_angle: "+cam_angle_Y);
}

void mouseReleased() {
    float[] rotations = cam.getRotations();
    xRotation = rotations[0];
    yRotation = rotations[1];
    zRotation = rotations[2];   
}

